# coding=UTF-8
from __future__ import absolute_import, division, print_function

"""
Definitions of framing.
Mechanisms for serialization/deserialization of AMQP framing and other types.


definitions.py is machine-generated from AMQP specification.
"""
