# coding=UTF-8
from __future__ import absolute_import, division, print_function

"""
- Compile XML to definitions.py
- Dynamically compile classes from property_flags
"""

from .content_property import compile_particular_content_property_list_class
